/* ======================================================================== */
/* cgolik.cpp                                                               */
/*	This file contains subroutines that calculate the Go-type model of  */
/* Hoang and Cieplak(JCP Vol. 112 Number 15 Page 6851 (2000)).              */
/*									    */
/* Written by Thomas Knotts	8 Mar 04				    */
/* Cosolvent interaction added by Nitin Rathore				    */
/* ======================================================================== */

/* ======================================================================== */
/* cbb_golik(int)                                                           */
/*	This subroutine calculates the backbone potential for the Go-type   */
/* model. Bonds.							    */
/* ======================================================================== */

#include "defines.h"

double cbb_golik (int ibox)
{

	int k = ibox;
	/* ================================================================== */
	/*                                                                    */
	/* Zero out the energy accumulator                                    */
	/*                                                                    */
	/* ================================================================== */
	double ebond = 0.0;

	double k1 = -golik[k].eps;
	double k2 = k1 * 100;
	/* ================================================================== */
	/*                                                                    */
	/* Calculate the energy and force associated with each bond.  The     */
	/* loop is over the number of bonds.                                  */
	/*                                                                    */
	/* ================================================================== */
	for(int i=0; i<bondN[k]; i++) {
		/* ----------------------------------------- */
		/* Assign the atoms involved in the bonding. */
		/* ----------------------------------------- */
		int ia = bond[k][i].a;
		int ib = bond[k][i].b;
		double req = bond[k][i].req;

		/* ----------------------------------------- */
		/* Calculate the distances between the atoms */
		/* involved in the bonding.                  */
		/* ----------------------------------------- */
		double dx = atom[k][ib].x - atom[k][ia].x;
		double dy = atom[k][ib].y - atom[k][ia].y;
		double dz = atom[k][ib].z - atom[k][ia].z;

		/* ----------------------------------------- */
		/* Calculate the length of the bond and then */
		/* the energy using the Go-type model.       */
		/* ----------------------------------------- */
		double drr = dx*dx + dy*dy + dz*dz;
		double dr  = sqrt(drr);
		double dri = 1.0 / dr;

		double rr = dr - req;
		double rr2= rr * rr;
		double fr1 = k1*rr*2.0;
		double fr2 = k2*rr*rr2*4.0;
		ebond += -rr*(fr1*0.5 + fr2*0.25);
		double fr = fr1 + fr2;
		/* ----------------------------------------- */
		/* Calculate the forces.                     */
		/* ----------------------------------------- */
		double fx = fr * dri * dx;
		double fy = fr * dri * dy;
		double fz = fr * dri * dz;

		ff[k][ia].x -= fx;
		ff[k][ia].y -= fy;
		ff[k][ia].z -= fz;
		ff[k][ib].x += fx;
		ff[k][ib].y += fy;
		ff[k][ib].z += fz;

	}

	return (ebond);

}

/* ======================================================================== */
/* cba_golik(int)                                                           */
/*	This subroutine calculates the backbone potential for the Go-type   */
/* model. Bends.							    */
/* ======================================================================== */
double cba_golik (int ibox)
{

	int k = ibox;
	/* ================================================================== */
	/* Zero out the energy accumulator                                    */
	/* ================================================================== */
	double ebend = 0.0;

	double k_theta = -10.0*golik[k].eps*20;  //10.0=20/2

	if(model == 2009)
		k_theta = -1400*golik[k].eps;


	/* ================================================================== */
	/* Calculate the energy and force associated with each bend.  The     */
	/* loop is over the number of bends.                                  */
	/* ================================================================== */

	for(int i=0; i<bendN[k]; i++) {
		/* ----------------------------------------- */
		/* Assign the atoms involved in the bending. */
		/* ----------------------------------------- */
		int ia = bend[k][i].a;
		int ib = bend[k][i].b;
		int ic = bend[k][i].c;

		/* ----------------------------------------- */
		/* Calculate the distances between the atoms */
		/* involved in the bending.                  */
		/* ----------------------------------------- */
		double dax = atom[k][ia].x - atom[k][ib].x;
		double day = atom[k][ia].y - atom[k][ib].y;
		double daz = atom[k][ia].z - atom[k][ib].z;
		double dcx = atom[k][ic].x - atom[k][ib].x;
		double dcy = atom[k][ic].y - atom[k][ib].y;
		double dcz = atom[k][ic].z - atom[k][ib].z;

		/* ----------------------------------------- */
		/* Calculate the angle of the bend and then  */
		/* the energy using the CHARMM force field.  */
		/* ----------------------------------------- */
		double drai = 1.0 / sqrt(dax*dax + day*day + daz*daz);
		double erax = dax * drai;
		double eray = day * drai;
		double eraz = daz * drai;
		double drci = 1.0 / sqrt(dcx*dcx + dcy*dcy + dcz*dcz);
		double ercx = dcx * drci;
		double ercy = dcy * drci;
		double ercz = dcz * drci;

		double cosphi = erax*ercx + eray*ercy + eraz*ercz;
		double phi = acos(cosphi);
		double isinphi = 1.0 / sqrt(1.0 - cosphi*cosphi);

		double rr = phi - bend[k][i].angeq;
		double fr = k_theta * rr * 2.0;
		ebend += -fr * rr * 0.5;

		double irasin = drai * isinphi * fr;
		double ircsin = drci * isinphi * fr;

		/* ----------------------------------------- */
		/* Calculate the forces.                     */
		/* ----------------------------------------- */
		double fxa = (erax * cosphi - ercx) * irasin;
		double fya = (eray * cosphi - ercy) * irasin;
		double fza = (eraz * cosphi - ercz) * irasin;
		double fxc = (ercx * cosphi - erax) * ircsin;
		double fyc = (ercy * cosphi - eray) * ircsin;
		double fzc = (ercz * cosphi - eraz) * ircsin;


		ff[k][ia].x += fxa;
		ff[k][ia].y += fya;
		ff[k][ia].z += fza;
		ff[k][ib].x += -(fxa + fxc);
		ff[k][ib].y += -(fya + fyc);
		ff[k][ib].z += -(fza + fzc);
		ff[k][ic].x += fxc;
		ff[k][ic].y += fyc;
		ff[k][ic].z += fzc;


	}

	return (ebend);

}

/* ======================================================================== */
/* cda_golik(int)                                                           */
/*	This subroutine calculates the backbone potential for the Go-type   */
/* model.  Dihedrals.							    */
/* ======================================================================== */

double cda_golik (int ibox)
{

	int k = ibox;

	/* ================================================================== */
	/* Zero out the energy accumulator                                    */
	/* ================================================================== */
	double etorsion = 0.0;

	double k_phi = 0.2 * golik[k].eps*20;


	if(model == 2009)
		k_phi = 28 * golik[k].eps;


	/* ================================================================== */
	/* Calculate the energy and force associated with each dihedral       */
	/* angle. The loop is over the number of dihedral angles.             */
	/* ================================================================== */

	for(int i=0; i<torsN[k]; i++) {
		/* ----------------------------------------- */
		/* Assign the atoms involved in the          */
		/* dihedral angle.                           */
		/* ----------------------------------------- */
		int ia = tors[k][i].a;
		int ib = tors[k][i].b;
		int ic = tors[k][i].c;
		int id = tors[k][i].d;

		/* ----------------------------------------- */
		/* Calculate the distances between the atoms */
		/* involved in the dihedral angle.           */
		/* ----------------------------------------- */
		double dabx = atom[k][ib].x - atom[k][ia].x;
		double daby = atom[k][ib].y - atom[k][ia].y;
		double dabz = atom[k][ib].z - atom[k][ia].z;
		double dbcx = atom[k][ic].x - atom[k][ib].x;
		double dbcy = atom[k][ic].y - atom[k][ib].y;
		double dbcz = atom[k][ic].z - atom[k][ib].z;
		double dcdx = atom[k][id].x - atom[k][ic].x;
		double dcdy = atom[k][id].y - atom[k][ic].y;
		double dcdz = atom[k][id].z - atom[k][ic].z;

		/* ----------------------------------------- */
		/* Calculate the bond lengths and unit       */
		/* vectors involved in the dihedral angle.   */
		/* ----------------------------------------- */
		double rab   = sqrt(dabx*dabx + daby*daby + dabz*dabz);
		double irab  = 1.0 / rab;
		double erabx = dabx * irab;
		double eraby = daby * irab;
		double erabz = dabz * irab;
		double rbc   = sqrt(dbcx*dbcx + dbcy*dbcy + dbcz*dbcz);
		double irbc  = 1.0 / rbc;
		double erbcx = dbcx * irbc;
		double erbcy = dbcy * irbc;
		double erbcz = dbcz * irbc;
		double rcd   = sqrt(dcdx*dcdx + dcdy*dcdy + dcdz*dcdz);
		double ircd  = 1.0 / rcd;
		double ercdx = dcdx * ircd;
		double ercdy = dcdy * ircd;
		double ercdz = dcdz * ircd;

		/* ----------------------------------------- */
		/* Calculate the cross and dot products      */
		/* between unit vectors and the bond angles. */
		/* ----------------------------------------- */
		double abbcx = eraby * erbcz - erabz * erbcy;
		double abbcy = erabz * erbcx - erabx * erbcz;
		double abbcz = erabx * erbcy - eraby * erbcx;
		double cosb  = -(erabx*erbcx + eraby*erbcy + erabz*erbcz);
		double isinb2 = 1.0 / (1.0 - cosb*cosb);
		double isinb  = sqrt(isinb2);

		double dccbx = ercdy * erbcz - ercdz * erbcy;
		double dccby = ercdz * erbcx - ercdx * erbcz;
		double dccbz = ercdx * erbcy - ercdy * erbcx;
		double cosc  = -(ercdx*erbcx + ercdy*erbcy + ercdz*erbcz);
		double isinc2 = 1.0 / (1.0 - cosc*cosc);
		double isinc  = sqrt(isinc2);


		/* ----------------------------------------- */
		/* Calculate the torsion/dihedral angle.     */
		/* ----------------------------------------- */
		double abcdx = -(abbcy * dccbz - abbcz * dccby);
		double abcdy = -(abbcz * dccbx - abbcx * dccbz);
		double abcdz = -(abbcx * dccby - abbcy * dccbx);
		double num = (abcdx*erbcx + abcdy*erbcy + abcdz*erbcz);
		double signum;
		if (num >= 0.0) signum = 1.0;
		else signum = -1.0;
		double costau = -(abbcx*dccbx + abbcy*dccby + abbcz*dccbz) * isinb * isinc;
		if (costau > 1.0) costau = 1.0;
		if (costau < -1.0) costau = -1.0;
		double tau = signum * acos(costau);

		/* -----------------------------------------	*/
		/* Assign the values for phi and psi		*/
		/* -----------------------------------------	*/
		if(tors[k][i].phitag ==1)		tors[k][i].phi = tau;
		else if (tors[k][i].psitag ==1) tors[k][i].psi = tau;
		else if (tors[k][i].thetag ==1) tors[k][i].theta = tau;
		/* ----------------------------------------- */
		/* Calculate the force and energy.           */
		/* ----------------------------------------- */

		double fr  =  k_phi*(sin(tau-tors[k][i].delphi[0]));
		etorsion  +=  k_phi*(1+cos(tau-tors[k][i].delphi[0]));

		/* ----------------------------------------- */
		/* Transform the forces to cartesian         */
		/* coordinates and accumulate.               */
		/* ----------------------------------------- */
		double fa    = -fr * irab * isinb2;
		double fax   = fa * abbcx;
		double fay   = fa * abbcy;
		double faz   = fa * abbcz;
		ff[k][ia].x += fax;
		ff[k][ia].y += fay;
		ff[k][ia].z += faz;

		double fb1   = fr * (rbc - rab*cosb) * irab *irbc * isinb2;
		double fb2   = fr * cosc * irbc * isinc2;
		ff[k][ib].x += (fb1 * abbcx + fb2 * dccbx);
		ff[k][ib].y += (fb1 * abbcy + fb2 * dccby);
		ff[k][ib].z += (fb1 * abbcz + fb2 * dccbz);

		double fc1   = fr * (rbc - rcd*cosc) * ircd *irbc *isinc2;
		double fc2   = fr * cosb * irbc * isinb2;
		double fcx   = fc1 * dccbx + fc2 * abbcx;
		double fcy   = fc1 * dccby + fc2 * abbcy;
		double fcz   = fc1 * dccbz + fc2 * abbcz;
		ff[k][ic].x += fcx;
		ff[k][ic].y += fcy;
		ff[k][ic].z += fcz;

		double fd    = -fr *ircd * isinc2;
		double fdx   = fd * dccbx;
		double fdy   = fd * dccby;
		double fdz   = fd * dccbz;
		ff[k][id].x += fdx;
		ff[k][id].y += fdy;
		ff[k][id].z += fdz;

	}
	return (etorsion);
}

/* ======================================================================== */
/* cnbnd_dnagolik(int, double*)                                             */
/*                                                                          */
/* This subroutine calculates the nonbonded native and non-native           */
/* energies of the Go-type model.                                           */
/*                                                                          */
/* Passed Parameters:                                                       */
/*			ibox:	    The box number                          */
/*			*energy:    A pointer to an array defined in        */
/*                                  forces() or force_long().  It contains  */
/*                                  three entries.  Two for unshifted and   */
/*                                  shifted vdw energies and one for        */
/*                                  coulombic energies.                     */
/*                                                                          */
/* ======================================================================== */
void cnbnd_dnagolik(int ibox, double *energy)
{

	int k = ibox;

	/* ================================================================== */
	/* Zero out the energy accumulator                                    */
	/* ================================================================== */
	double enonbond	= 0.0;
	double enonbonds	= 0.0;
	double enonnative	= 0.0;
	double ecoulomb	= 0.0;
	double ebp		= 0.0;
	double emm		= 0.0;
	double esolv          = 0.0;
	/* ================================================================== */
	/* Set the parameters for the types of iteractions.                   */
	/* ================================================================== */
	double epsilon = golik[k].eps;

	double eps_AT, eps_GC;
	if(model == 2009)
	{
		eps_AT = 2.0*epsilon;
		eps_GC = 2.532*epsilon;
	}
	else // model is 2007
	{
		eps_AT = 4.0*epsilon*2.0/3.0;
		eps_GC = 4.0*epsilon;
	}

	double factor = pow(2.0,-(1.0/6.0));
	double sig_AT = 2.9002;
	double sig_GC = 2.8694;

	double d_mm   = 1.0;
	double sig_mm = factor*d_mm;
	double eps_mm = epsilon;

	/* ================================================================== */
	/*                                                                    */
	/* Get and assign information for pbc and cutoffs                     */
	/*                                                                    */
	/* ================================================================== */

	double rc2  = box[k].rc2;
	double rc2i = 1.0/box[k].rc2;

	double rc_mm = d_mm;
	double rc2_mm = rc_mm * rc_mm;


	double eps_inf=0, A_I=0;
	if(model == 2009)
	{
		eps_inf = 0.504982*epsilon;
		A_I = 0.474876*(1.0 + 1.0 / (0.148378 + 10.9553*sim.Na_con[k]));
	}

	/* ================================================================== */
	/*                                                                    */
	/* Calculate the energy and force on each molecule due to atoms in    */
	/* its neighborlist. The first loop is over sites/atoms 0 to N-1      */
	/* and the second is over the number of neighbors in atom a's         */
	/* neighborlist.                                                      */
	/*                                                                    */
	/* ================================================================== */

	for(int a=0; a<box[k].boxns-1; a++) {

		/* ----------------------------------------------- */
		/* Loop around neighbors of particle a.            */
		/* ----------------------------------------------- */
		for(int bb=0; bb<nlist[k].count[a]; bb++) {
			int b = nlist[k].list[a][bb];
			/* ----------------------------------------------- */
			/* Calculate the x,y,z distances between           */
			/* particles a & b.                                */
			/* ----------------------------------------------- */
			double dx = atom[k][a].x - atom[k][b].x;
			double dy = atom[k][a].y - atom[k][b].y;
			double dz = atom[k][a].z - atom[k][b].z;

			/* ----------------------------------------------- */
			/* Calculate the distance between particles a & b. */
			/* If the distance is less than the cutoff,        */
			/* calculated the energy and the force. If it is   */
			/* greater, go to the next pair in the loop.       */
			/* ----------------------------------------------- */
			double dr2  = dx*dx + dy*dy + dz*dz;
			double dr2i = 1.0/dr2;

			if(dr2 > rc2) continue;

			/* ----------------------------------------------- */
			/* Set the epsilon, sigma, and qq for the pair.    */
			/* ----------------------------------------------- */
			double sig, eps, eps4;
			double sr2,

			       sr6, sr10, sr12;
			double sr2s, sr6s, sr10s, sr12s;

			int Nres_a;
			int Nres_b;
			int N_bp;
			double eps_N;
			double eps_s;
			double s_exp;
			double s_factor2;

			double fx, fy, fz;
			int xflag=nlist[k].xflag[a][bb];
			int ta = atom[k][a].site_id;
			int tb = atom[k][b].site_id;

			sig     = ljset[ta][tb].sig;
			double rc2_non = ljset[ta][tb].rc2_non;
			int flag = ljset[ta][tb].id;

			if(xflag){  //flag is 0 if there is no exclusion
				if(xflag == 1) flag = 1;
				else if(xflag == 5) flag = 5;
				else if(xflag == 6) flag = 6;
				else if(xflag == 8) flag = 0;
			}

			/* ----------------------------------------------- */
			/* Calculate the LJ force and energy.              */
			/* ----------------------------------------------- */
			double force = 0.0; //double enbnd =0.0; double enbnds = 0.0;

			switch(flag){
				case 1: //native
					sig = nlist[k].sig[a][bb];
					eps4 = 4.0*epsilon;
					sr2  = (sig * sig) * dr2i;
					sr6  = sr2 * sr2 * sr2;
					sr12 = sr6 * sr6;

					sr2s  = (sig * sig) * rc2i;
					sr6s  = sr2s * sr2s * sr2s;
					sr12s = sr6s * sr6s;

					force      = (12.0*eps4*dr2i) * (sr12 - 0.5*sr6);
					enonbond  += eps4 * (sr12 - sr6);
					enonbonds += eps4 * (sr12s -sr6s);

					break;
				case 2: //AT compliment
					sig = sig_AT;
					eps4 = 4.0*eps_AT;
					sr2  = (sig * sig) * dr2i;
					sr6  = sr2 * sr2 * sr2;
					sr12 = sr6 * sr6;
					sr10 = sr12/sr2;

					sr2s  = (sig * sig) * rc2i;
					sr6s  = sr2s * sr2s * sr2s;
					sr12s = sr6s * sr6s;
					sr10s = sr12s/sr2s;

					force      = (60.0*eps4*dr2i) * (sr12 - sr10);
					ebp  += eps4 * (5.0*sr12 - 6.0*sr10);
					ebp  -= eps4 * (5.0*sr12s- 6.0*sr10s);

					break;
				case 3: //CG compliment
					sig = sig_GC;
					eps4 = 4.0*eps_GC;
					sr2  = (sig * sig) * dr2i;
					sr6  = sr2 * sr2 * sr2;
					sr12 = sr6 * sr6;
					sr10 = sr12/sr2;

					sr2s  = (sig * sig) * rc2i;
					sr6s  = sr2s * sr2s * sr2s;
					sr12s = sr6s * sr6s;
					sr10s = sr12s/sr2s;

					force      = (60.0*eps4*dr2i) * (sr12 - sr10);
					ebp  += eps4 * (5.0*sr12 - 6.0*sr10);
					ebp  -= eps4 * (5.0*sr12s- 6.0*sr10s);

					break;
				case 4: //mis-match


					if(dr2<rc2_mm){
						sig = sig_mm;
						eps = eps_mm;
						sr2  = (sig * sig) * dr2i;
						sr6  = sr2 * sr2 * sr2;
						sr12 = sr6 * sr6;

						force        = (48.0*eps*dr2i) * (sr12 - 0.5*sr6);
						enonnative  += 4.0*eps*(sr12 - sr6) + eps;
					}

					break;
				case 5://bonded pairs
					break;
				case 6://native that is also a bend
					sig = nlist[k].sig[a][bb];
					eps4 = 4.0*epsilon;
					sr2  = (sig * sig) * dr2i;
					sr6  = sr2 * sr2 * sr2;
					sr12 = sr6 * sr6;

					sr2s  = (sig * sig) * rc2i;
					sr6s  = sr2s * sr2s * sr2s;
					sr12s = sr6s * sr6s;

					force      = (12.0*eps4*dr2i) * (sr12 - 0.5*sr6);
					enonbond  += eps4 * (sr12 - sr6);
					enonbonds += eps4 * (sr12s -sr6s);

					break;

				case 7://attraction between sugars on different strands
					//For finding the number of residues on the shorter strand
					if(model == 2009)
					{
						Nres_a = mol[residue[k][atom[k][a].residue].chain].Nres;
						Nres_b = mol[residue[k][atom[k][b].residue].chain].Nres;
						if (Nres_a < Nres_b) N_bp = Nres_a;
						else N_bp = Nres_b;

						eps_N = eps_inf * (1.0 - 1.0 / (1.40418 - 0.268231*N_bp));
						eps_s = A_I * eps_N;
						s_exp = exp(-0.1875*(sqrt(dr2) - 13.38));//alpha_s = 0.1875, r_s = 13.38
						s_factor2 = (1 - s_exp)*(1 - s_exp);

						force = -2.0 * eps_s * 0.1875 * s_exp * (1 / sqrt(dr2)) * (1.0 - s_exp);//alpha_s = 0.1875
						esolv += eps_s * s_factor2 - eps_s;

						break;
					}
					else // model is 2007
					{
						//The correct behavior is to pass into the default block below
					}

				default://non-native

					if(dr2<rc2_non){
						//sig = sig_non;
						eps = epsilon;
						sr2  = (sig * sig) * dr2i;
						sr6  = sr2 * sr2 * sr2;
						sr12 = sr6 * sr6;

						force        = (48.0*eps*dr2i) * (sr12 - 0.5*sr6);
						enonnative  += 4.0*eps*(sr12 - sr6) + eps;
					}
					break;
			}

			/* ----------------------------------------------- */
			/* Calculate the Coulombic energy and forces.      */
			/* ----------------------------------------------- */


			double qq = pott[k][a].qq * pott[k][b].qq * QFACTOR;

			if(qq != 0.0){

				switch(flag){
					case 5: //bond


						break;
					case 6: //bend


						break;
					default:

						qq = qq / sim.epsRF[k];
						double lambdai = 1.0/sim.kappa[k];
						double dri  = sqrt(dr2i);
						double qqerlambda = qq*exp(-lambdai/dri)*dri;
						force += qqerlambda*(dri*lambdai+dr2i);
						ecoulomb += qqerlambda;


				}
			}

			/* ----------------------------------------- */
			/* Accumulate the forces.                    */
			/* ----------------------------------------- */
			if(force != 0.0){
				fx	    = force * dx;
				ff[k][a].x += fx;
				ff[k][b].x -= fx;

				fy	    = force * dy;
				ff[k][a].y += fy;
				ff[k][b].y -= fy;

				fz	    = force * dz;
				ff[k][a].z += fz;
				ff[k][b].z -= fz;


			}//if(force!=0.0)

		}//loop over number of neighbors of molecule a

	}//loop over molecules a

	/* ==================================================================== */
	/*                                                                      */
	/* After the loop, assign the accumulated energies to "energy" array.   */
	/*	energy[0] = enonbond						*/
	/*	energy[1] = enonbond shifted					*/
	/*	energy[2] = EPP protein-protein nonbond shifted			*/
	/*	energy[3] = EPS protein-solute nonbond shifted			*/
	/*	energy[4] = ESS solute-solute nonbond shifted			*/
	/*                                                                      */
	/* ==================================================================== */
	energy[0] = enonbond+enonnative+ebp+emm+ecoulomb+esolv;          // non-shifted energy (LJ and coulomibic)
	energy[1] = enonbond-enonbonds+enonnative+ebp+emm+ecoulomb+esolv; // shifted energy (LJ and coulomibic) + enonnative
	energy[2] = enonbond-enonbonds;
	energy[3] = enonnative+emm;
	energy[4] = ebp;
	energy[5] = ecoulomb;
	energy[11]= esolv;

}

