/* ======================================================================== */
/* defines.h                                                                */
/*                                                                          */
/* This header file contains the constants, structures, and libraries       */
/* used in the program.  The structures defined here are global variables   */
/* accessible to any subroutine where #include defines.h is put at the top  */
/* of the file.                                                             */
/*                                                                          */
/* Note: To make a variable global in this manner, #ifdef MAIN is used to   */
/* include the key word "extern" before the structure is defined.  MAIN is  */
/* only defined in the subroutine main(). Thereafter, extern tells the      */
/* compiler that this variable has been declared elsewhere, but to make it  */
/* accessible to the current subroutine.                                    */
/* ======================================================================== */

/* ================================================================== */
/*                                                                    */
/* Libraries and Constants                                            */
/*                                                                    */
/* ================================================================== */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>

#define PI      3.141592653589793238462643383		/* pi			                   */
#define KB      1.380662E-23				/* Boltzmann constant [J/K]	           */
#define NA      6.022045E23				/* Avogadro  constant [/mol]	           */
#define RG      (KB*NA)
#define AU      1.6605655E-27				/* atomic massunit    [kg]           	   */
#define ELEC    8.987551789e19				/* [J.A/C^2] (4.Pi.e)^-1                   */
#define ELEQ    1.602176462e-19				/* [C]                                     */
#define QFACTOR ELEC*ELEQ*ELEQ*NA*1.0E-3                /* conversion factor for charges to kJ/mol */

#define NBOXES  1					/* max number of boxes                     */
#define NTYPES  8					/* max type of molecules                   */
#define MEXTRA  100					/* size for extra memory allocation        */
#define NPRIME  1009					/* number of entries in the hash table     */
#define NUM_LJSET	300				/* This number controls the size of ljset  */

#define greater(A,B) ((A)>(B) ? (A):(B))
#define ROUND(q, p)   (((int)((q)*pow(10.0,(p))+0.5*((q)>0?1:-1)))/(pow(10.0,(p))))

extern int model;

/* ================================================================== */
/*                                                                    */
/* Strutures of Molecular Variables                                   */
/*                                                                    */
/* ================================================================== */
/* ----------------------------------------- */
/* Globals for number of atoms types, bonds  */
/* bends, ureys, torsions, impropers that    */
/* are read in read_atom from the param file */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
char param_file_name[50];

#ifndef MAIN
extern
#endif
int NUM_ATOM;

#ifndef MAIN
extern
#endif
int NUM_BOND;

#ifndef MAIN
extern
#endif
int NUM_BEND;

#ifndef MAIN
extern
#endif
int NUM_TORS;

#ifndef MAIN
extern
#endif
int NUM_IMPR;

#ifndef MAIN
extern
#endif
int NUM_UREY;

#ifndef MAIN
extern
#endif
int NUM_NBFIX;

/* ----------------------------------------- */
/* atoms                                     */
/* Contains information about each site.     */
/* The site symbol, name, position, charge   */
/* molecule number, and id number are here.  */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct atoms {
	char   t[5];          /* symbol for site           */
	char   site_name[5];
	char	res_name[4];
	double mass;
	double x;             /* x-position of site        */
	double y;             /* y-position of site        */
	double z;             /* z-position of site        */
	double charge;             /* site charge               */
	int residue;                /* index for residue number  */
	int molec;            /* index for site number     */
	int site_id;						
} **atom;

/* ----------------------------------------- */
/* res                                       */
/* Contains information about residue: the   */
/* residue type, number of sites in the      */
/* residue, and the number of chains.        */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct res {
	int type;
	int Nsite;
	int chain;  
} **residue;

/* ----------------------------------------- */
/* mol_info                                  */
/* Contains information about each type of   */
/* molecule: total number of molecule i,     */
/* the number of sites in molecule i, the    */
/* number of residues in molecule i, and the */
/* molecular weight of molecule i.           */
/* ----------------------------------------- */
#ifndef MAIN
extern 
#endif
struct mol_info {
	int    Ni;                           /* total number of molecules i      */
	int    Nsite;                        /* number of sites of molecule i    */
	int	 Nres;                         /* number of residues in molecule i */
	double MW;                           /* molecular weight of molecule i   */
} *mol;

/* ----------------------------------------- */
/* mol_id                                    */
/* Contains information about each specific  */
/* molecule in the system.  It gives the     */
/* beginning site number of each molecule    */
/* and the number of molecules of interest   */
/* to perform special moves on, e.g. the     */
/* proteins in the system rather than the    */
/* water might want to have a pivot move.    */
/* ----------------------------------------- */
#ifndef MAIN
extern 
#endif
struct mol_id {
	int   *fsite;			/* first site number of molecule i                 */
	int   *lsite;			/* last site number of molecule i                  */
	int   *fres;			/* beginning residue number of each molecule i     */
	int   *lres;			/* last residue number in each molecule            */
	int   Nsolute;		/* Number of molecules to perform fancy MC move on */

} molec;

/* ----------------------------------------- */
/* Nmolall = Total number of molecules in    */
/*       all the boxes.                      */
/* Nsitall = Total number of sites in all    */
/*       the boxes.                          */
/* ----------------------------------------- */

#ifndef MAIN
extern
#endif
int Nmolall, Nsitall;

/* ----------------------------------------- */
/* nlist                                     */
/* The neighborlist structure. The index on  */
/* count runs over all sites.  The indicies  */
/* on list run over all sites and all        */
/* neighbors.                                */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct nlist_def{
	int *count;
	int **list;
	int **xflag;
	double **sig;

} *nlist;

#ifndef MAIN
extern
#endif
int max_num_neighbors;

/* ----------------------------------------- */
/* veloc                                     */
/* This structure contains the x,y,z,        */
/* components of the velocities and forces.  */
/* ff_short and ff_long are used for mts     */
/* integration.                              */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif 
struct veloc {
	double x;
	double y;
	double z;
} **ff;

/* ----------------------------------------- */
/* energy                                    */
/* This structure contains the information   */
/* on all the energy of the system.  Both    */
/* kinetic and potential energy are          */
/* included as well as the contributions to  */
/* the potential energy and the extended     */
/* hamiltonian for Nose Hoover dynamics.     */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct energy {
	double poten;
	double potens;
	double nbond;
	double nbonds;
	double bond;
	double bend;
	double tors;
	double impr;
	double coulomb;
	double dna_stack;
	double dna_nonnat;
	double dna_bp;
	double dna_solv;
} en[NBOXES];

/* ----------------------------------------- */
/* inter                                     */
/* This structure contains the information   */
/* on the LJ parameters, mass, and charge    */
/* of each site.  The indices on the arrays  */
/* run over the box number and the site      */
/* number.                                   */
/* ----------------------------------------- */

#ifndef MAIN
extern
#endif
struct inter {
	double mas;                   /* mass of each atom/site              */
	double sig;                   /* sigma parameter for LJ potential    */
	double eps;                   /* epsilon parameter for LJ potential  */
	double qq;                    /* site charge for Coulombic potential */  
	int    id;                    /* symbol for atom/site                */ 
	double rc2_non;
} **pott, **ljset;

/* ================================================================== */
/*                                                                    */
/* The following 9 structures contain the information used to         */
/* calculate the energies and forces on the system. They are assigned */
/* by comparing the atoms/sites involved in system with the different */
/* read in read_atom from the simul22.param file.  The sturctures     */
/* that contain the infor from the simul22.param file are the 5       */
/* just below the 9 here.                                             */
/*                                                                    */
/* ================================================================== */

/* ----------------------------------------- */
/* bonds                                     */
/* This structure contains the information   */
/* on each bond in the system.  The site     */
/* number of the atoms involved in the bond  */
/* and the force field parameters are here.  */
/* the indicies run over the number of       */
/* boxes and the number of bonds.            */
/* Values are set in read_topology.          */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct bonds {
	int a;
	int b;
	double req;
	double krbond;
} **bond;

/* ----------------------------------------- */
/* bondN[k] = Total number of bonds in box k */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif 
int bondN[NBOXES];

/* ----------------------------------------- */
/* bends                                     */
/* This structure contains the information   */
/* on each bend in the system.  The site     */
/* number of the atoms involved in the bend  */
/* and the force field parameters are here.  */
/* the indicies run over the number of       */
/* boxes and the number of bends.            */
/* Values are set in read_topology.          */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct bends {
	int a;
	int b;
	int c;
	double angeq;
	double krbend;
} **bend;

/* ----------------------------------------- */
/* bendN[k] = Total number of bends in box k */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
int bendN[NBOXES];

/* ----------------------------------------- */
/* torsions                                  */
/* This structure contains the information   */
/* on each torsion in the system.  The site  */
/* number of the atoms involved in the       */
/* dihedral and the force field parameters   */
/* are here. The indicies run over the       */
/* number of boxes and the number of         */
/* torsions.                                 */
/* Values are set in read_topology.          */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct torsions {
	int a;
	int b;
	int c;
	int d;
	int count;
	double phi;
	double psi;
	double theta;
	double phia;
	double psia;
	double thetaa;
	char phitag;
	char psitag;
	char thetag;
	double kphi[6];
	double nphi[6];
	double delphi[6];
} **tors;

/* ----------------------------------------- */
/* torsN[k] = Total number of torsions in    */
/* box k.                                    */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
int torsN[NBOXES];

/* ----------------------------------------- */
/* excinter                                  */
/* This structure contains the information   */
/* on each modified interaction in the       */
/* system.  Note: These interactions are not */
/* used in the current version.              */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct excinter {
	int a;
	int b;
	int xflag;
	double eps;
	double sig;
	double qa;
	double qb;
} **xint;

/* ----------------------------------------- */
/* xintN[k] = Total number of modified       */
/* interactions in box k.                    */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
int xintN[NBOXES];

/* ----------------------------------------- */
/* bondprop                                  */
/* This structure contains the information   */
/* on each type of bond in the system as     */
/* read in read_atom() from the              */
/* simul22.param file.                       */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct bondprop {
	int type1;
	int type2;
	double kbond;
	double eqbond;
} *bond_prop;

/* ----------------------------------------- */
/* ureyprop                                  */
/* This structure contains the information   */
/* on each type of bond in the system as     */
/* read in read_atom() from the              */
/* simul22.param file.                       */
/* ----------------------------------------- */

#ifndef MAIN
extern
#endif
struct ureyprop {
	int type1;
	int type2;
	int type3;
	double k;
	double Seq;
} *urey_prop;

/* ----------------------------------------- */
/* bendprop                                  */
/* This structure contains the information   */
/* on each type of bend in the system as     */
/* read in read_atom() from the              */
/* simul22.param file.                       */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct bendprop {
	int type1;
	int type2;
	int type3;
	double kbend;
	double eqbend;
}*bend_prop;

/* ----------------------------------------- */
/* dihedralprop                              */
/* This structure contains the information   */
/* on each type of torsion in the system as  */
/* read in read_atom() from the              */
/* simul22.param file.                       */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct dihedralprop {
	int type1;
	int type2;
	int type3;
	int type4;
	double kphi;
	double nphi;
	double delphi;
} *dihedral_prop;

/* ----------------------------------------- */
/* improperprop                              */
/* This structure contains the information   */
/* on each type of improper torsion in the   */
/* system as read in read_atom() from the    */
/* simul22.param file.                       */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct improperprop {
	int type1;
	int type2;
	int type3;
	int type4;
	double kimpr;
	double angeq;
} *improper_prop;

#ifndef MAIN
extern
#endif
struct nbfixprop{
	int type1;
	int type2;
	double sig;                       /* sigma parameter for LJ potential    */
	double eps;                       /* epsilon parameter for LJ potential  */
}*nbfix_prop;

/* ----------------------------------------- */
/* epstmp                                    */
/* This structure contains the epsilon value */
/* of the LJ potential for each type of site */
/* as read in reac_atom() from the           */
/* simul22.param file.                       */
/* The first index:                          */
/*   0 = Regular NBOND                       */
/*   1 = Special 14 NBOND                    */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
double epstmp[2][2000];

/* ----------------------------------------- */
/* sigtmp                                    */
/* This structure contains the sigma value   */
/* of the LJ potential for each type of site */
/* as read in reac_atom() from the           */
/* simul22.param file.                       */
/* The first index:                          */
/*   0 = Regular NBOND                       */
/*   1 = Special 14 NBOND                    */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
double sigtmp[2][2000];

/* ----------------------------------------- */
/* atomtype                                  */
/* This structure contains each type of      */
/* atom/site read in read_atom.              */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct atomtype {
	char name[5];
	int atomid;
	double mass;
} *atom_type;

/* ----------------------------------------- */
/* SimData                                   */
/* This structure contains information on    */
/* the simulation as read in read sim.       */
/* ----------------------------------------- */
#ifndef MAIN
extern 
#endif
struct SimData {
	int    NB;                   /* number of simulation boxes           */
	int    NC;                   /* number of components                 */
	double T[NBOXES];            /* temperature [K]                      */
	double kT[NBOXES];           /* kT = KB * T                          */
	double rc;                   /* cutoff radius [A]                    */
	double rclist;               /* cutoff radius for neighbor-list[A]   */
	double rc_ew;                /* cutoff radius for ewald [A]          */
	double rclist_ew;            /* cutoff radius for ewald nlist [A]    */
	double kappa[NBOXES];        /* Ewald parameter for Gaussian width   */
	double Na_con[NBOXES];       /* Sodium concentration                 */
	double epsRF[NBOXES];        /* Dielectric constant                  */
} sim;

/* ----------------------------------------- */
/* BoxData                                   */
/* This structure contains information on    */
/* each box.                                 */
/* ----------------------------------------- */
#ifndef MAIN
extern 
#endif
struct BoxData {
	int    boxn;			/* total number of molecules in box i     */
	int    boxns;			/* total number of sites in box i         */
	int    boxnres;			/* total number of residues in box i      */
	double rc2;			/* cutoff radius (squared)                */
	double rc4;			/* cutoff radius ^4                       */
	double weight;			/* total weight                           */
	double temp;			/* instantaneous temperature              */
} box[NBOXES];

/* ----------------------------------------- */
/* BoxProp                                   */
/* This structure contains information on    */
/* number of different molecules in each     */
/* box.                                      */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct BoxProp {
	int first;
	int last;
	int nbox;
} bp[NBOXES][NTYPES];

/* ----------------------------------------- */
/* This structure contains information about */
/* the hash table where excluded interaction */
/* are stored.                               */
/* ----------------------------------------- */
#ifndef MAIN
extern
#endif
struct hashlist{
	int *a, *b;	/* Site numbers involved in the exluded interction           */
	int n;		/* Number of occupants at a specific index of the hash table */
	int *p;		/* Identifies which i of xint[k][i] is at this position      */
} **table;
#ifndef MAIN
extern
#endif
int hashsize;

#ifndef MAIN
extern
#endif
struct golik_struct{
	double eps;
	double req;
	double d_native;
	double d_repulsive;

	double *hydro_index;
	double mass;
	int Psite;	// Site index in crd where solute starts for energy partition
}*golik;

extern int model;

#ifndef MAIN
extern
#endif

struct{
	double x;
	double y;
	double z;
	double r;
	double theta;
	double theta_2; //angle on opposite strand
	double mass;
	double charge;
	int site_id;
}PPP,SSS,AAA,GGG,TTT,CCC;

#ifndef MAIN
extern
#endif
struct{
	int N;
	char* base;
}sequence;

#ifndef MAIN
extern
#endif
struct atoms_dna{
	double x;
	double y;
	double z;
	double mass;
	double charge;
	char site_name[4];
	char res_name[4];
	int residue;
	int site_id;
}*atom_dna;

#ifndef MAIN
extern
#endif
struct atomN_struct{
	int tot; //total number of atoms
	int Nsalt; //total number of NaCl molecules
	int dna1; //site where dna1 begins
	int dna2; //site where dna2 begins
	int sod; //site where condensed sod counter ions begins
} natom_dna;

#ifndef MAIN
extern
#endif
int bondN_dna;

#ifndef MAIN
extern
#endif
struct bonds_dna{
	int a;
	int b;
}*bond_dna;

#ifndef MAIN
extern
#endif
int bendN_dna;

#ifndef MAIN
extern
#endif

struct bends_dna{
	int a;
	int b;
	int c;
}*bend_dna;

#ifndef MAIN
extern
#endif
int torsN_dna;

#ifndef MAIN
extern
#endif

struct torss_dna{
	int a;
	int b;
	int c;
	int d;
}*tors_dna;

#ifndef MAIN
extern
#endif
char fileprefix[256];

