/********************************************************************************************/
/***                       Coarse Grain DNA Force Field Code                              ***/
/***    by Thomas A Knotts IV, Nitin Rathore, David C. Scwhartz, and Juan J. de Pablo     ***/
/***  Knotts et al., "A Coarse-grain Model for DNA," J. Chem. Phys., 126, 084901 (2007).  ***/
/********************************************************************************************/

#define MAIN
#include "defines.h"
/* ================================================================== */
/*                                                                    */
/* Function prototypes                                                */
/*                                                                    */
/* ================================================================== */

void init_all     (void);
void init_param	  (void);
void read_atom	  (void);
void read_config  (void);
void setlj		  (void);
void weight       (void);
void forces_format(int);
void calcvalue	  (int);
void nblist		  (int);
void read_dnagolik(void);
int dna2go(int argc, char* argv[]);

int model;

/* ************************************************************************ */
/*                                                                          */
/* ====================== Begin main Subroutine =========================== */
/*                                                                          */
/* ************************************************************************ */
int main(int argc, char* argv[])
{
	printf("\n");
	printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	printf("@@                     Coarse Grain DNA Force Field Code                     @@\n");
	printf("@@        by Thomas A Knotts IV, Jacob I Lewis, and Eric R Mansfield         @@\n");
	printf("@@                   Brigham Young University, Provo, Utah                   @@\n");
	printf("@@                                                                           @@\n");
	printf("@@  Knotts et al., \"A Coarse-grain Model for DNA,\" J. Chem. Phys. 126,       @@\n");
	printf("@@      084901 (2007).                                                       @@\n");
	printf("@@  Sambriski et al., \"Sequence effects in the melting and renaturation of   @@\n");
	printf("@@      short DNA oligonucleotides: structure and mechanistic pathways,\"     @@\n");
	printf("@@      J. Phys.: Condens. Matter 21, 034105 (2009).                         @@\n");
	printf("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
	int usage = 1;
	// Read in the arguments

	if(argc == 5)
		if(sscanf(argv[2], "%lf", &sim.T[0])==1)
			if(sscanf(argv[3], "%lf", &sim.Na_con[0]) == 1)
				if(sscanf(argv[4], "%d", &model))
					if(model == 2007 || model == 2009)
						usage = 0;

    if(usage == 1)
	{
		printf("\nUsage:\n");
		printf("  %s filename t n y\n", argv[0]);
		printf("\n");
		printf("filename = Name of file containing a single DNA sequence\n");
		printf("t        = Temperature in K\n");
		printf("n        = Salt (Na+) concentration in mM\n");
		printf("y        = the model version, either 2007 or 2009\n");
		return 1;
	}

	strcpy(fileprefix, argv[1]);

	sim.kT[0] = RG * sim.T[0] * 0.001; // This is modified to get the units in KJ/mol
	sim.Na_con[0] *= 0.001;
	strcpy(param_file_name, "coarse_grain.param");

	if(model == 2009)
	{
		//The simulation, in read_sim.c, reads Na_con and calculates the kappa and eps_rf.

		sim.epsRF[0] =(249.4-0.788*sim.T[0]+7.2E-4*sim.T[0]*sim.T[0]) * (1.0-0.2551*sim.Na_con[0]+0.05151*sim.Na_con[0]*sim.Na_con[0]-6.889E-3*sim.Na_con[0]*sim.Na_con[0]*sim.Na_con[0]);
		if(sim.Na_con[0]==0)
			sim.kappa[0] = 78;
		else
			sim.kappa[0] = sqrt((sim.epsRF[0] * sim.kT[0]) / (8 * PI * ELEC * NA * NA * ELEQ * ELEQ * sim.Na_con[0]) * 1.0E30);

	}
	else // model is 2007
	{
		//The simulation reads in kappa and eps_rf. In our case however, use the Na_con to calculate those.
		double Q0 = 8.85419e-12;
		double QR = 78;	// Assumed in 2007 model
		double R = 8.3144090;
		sim.epsRF[0] = 78.0;
		sim.kappa[0] = sqrt((Q0 * QR * R * sim.T[0]) / (2*NA*NA*ELEQ*ELEQ*greater(sim.Na_con[0],1e-7))*1e17);
	}


	// Generate DNA
	dna2go(argc, argv);

	if(sequence.N > 2000)
		printf("Note: Large sequences may take a while!\n");


	/* ----------------------------------------------- */
	/* Get the overall simulation parameters from the  */
	/* simul.input file.                               */
	/* ----------------------------------------------- */
	sim.NB = 1;
	sim.rc = 50;
	sim.rclist = 52;

	/* ----------------------------------------------- */
	/* Initialize all the property variables to zero   */
	/* ----------------------------------------------- */
	init_all();

	/* ----------------------------------------------- */
	/* Read in the force field parameters from the     */
	/* simul19.input file.                             */
	/* ----------------------------------------------- */
	read_atom();

	/* ----------------------------------------------- */
	/* Read in the x,y,z coordinates of the sites.     */
	/* ----------------------------------------------- */
	read_config();

	/* ----------------------------------------------- */
	/* Calculate the LJ parameters for each site/site  */
	/* interaction.                                    */
	/* ----------------------------------------------- */
	setlj();

	/* ----------------------------------------------- */
	/* Initialize all the property variables to zero   */
	/* ----------------------------------------------- */
	init_param();

	/* ----------------------------------------------- */
	/* Read in Go-like model parameters.               */
	/* ----------------------------------------------- */
	read_dnagolik();

	/* ----------------------------------------------- */
	/* Calculate the neighborlist, forces, energies,   */
	/* and other properties.                           */
	/* ----------------------------------------------- */
	nblist(0);
	forces_format(0);
	calcvalue(0);

	free(nbfix_prop);

	// Output the energies in a nice format
	printf("\n");
	time_t now;
	time(&now);
	printf("Date and Time: %.24s \n",ctime(&now));
	printf("Parameter file name: %s \n", param_file_name);
	printf("\n");
	printf("Sequence (%d bp):\n", sequence.N);

	int i;
	for(i=0; i<sequence.N; i++)
	{
		if(i%50==0 && i>0)
			printf("\n");
		else if(i%5==0 && i>0)
			printf(" ");

		printf("%c", sequence.base[i]);
	}
	printf("\n\n");
	printf("Model used: %d\n", model);
	printf("Temperature:            %12.5f K\n", sim.T[0]);
	printf("Salt concentration:     %12.5f M\n", sim.Na_con[0]);
	if(model == 2007)
	{
		printf("Bond energy:            %12.5f kJ/mol\n", en[0].bond);
		printf("Angle energy:           %12.5f kJ/mol\n", en[0].bend);
		printf("Dihedral energy:        %12.5f kJ/mol\n", en[0].tors);
		printf("Stacking energy:        %12.5f kJ/mol\n", en[0].dna_stack);
		printf("Base pairing energy:    %12.5f kJ/mol\n", en[0].dna_bp);
		printf("Excluded volume energy: %12.5f kJ/mol\n", en[0].dna_nonnat);
		printf("Coulombic energy:       %12.5f kJ/mol\n", en[0].coulomb);
		printf("Solvent-induced energy: %12.5f kJ/mol\n", en[0].dna_solv);
		printf("Total potential energy: %12.5f kJ/mol\n", en[0].potens);
	}
	else
	{
		printf("Bond energy:            %12.5f kJ/mol\n", en[0].bond);
		printf("Bend energy:            %12.5f kJ/mol\n", en[0].bend);
		printf("Torsion energy:         %12.5f kJ/mol\n", en[0].tors);
		printf("Stacking energy:        %12.5f kJ/mol\n", en[0].dna_stack);
		printf("Base pairing energy:    %12.5f kJ/mol\n", en[0].dna_bp);
		printf("Non-native energy:      %12.5f kJ/mol\n", en[0].dna_nonnat);
		printf("Electrostatic energy:   %12.5f kJ/mol\n", en[0].coulomb);
		printf("Solvent-induced energy: %12.5f kJ/mol\n", en[0].dna_solv);
		printf("Total potential energy: %12.5f kJ/mol\n", en[0].potens);
	}

	printf("\n");
	printf("Outputted files:\n");
	printf("%s.pdb\n", fileprefix);
	printf("%s.psf\n", fileprefix);
	printf("%s.crd\n", fileprefix);
	printf("%s.allatom.pdb\n", fileprefix);

	return 0;
}

