/* ======================================================================== */
/* read_atom.cpp                                                            */
/*                                                                          */
/*   This subroutine reads the atom types and parameters from the           */
/* database.  It reads the information contained in simul19.param if        */
/* NEUTRAL is not defined and simul19_sasa.param if NEUTRAL is defined.     */
/* This subroutine gets the information about the atom/site id number,      */
/* name, and molecular weight/mass.  It then reads in the bond information  */
/* for all possible bonds, followed by the bending, torision, and LJ        */
/* information.                                                             */  
/* 									    */	
/*  	                                                                    */
/* Note:  This subroutine basically just reads in the CHARMM force field    */
/* parameters. All of them are not necessarily used in a simulations. The   */
/* atoms, bonds, bends, torsions, and LJ parameters used are read in        */
/* read_topology and setlj.  In these subroutines, the each atom/structure  */
/* present in the simulation is compared to those in this list to get the   */
/* appropriate value.                                                       */
/*                                                                          */
/* ======================================================================== */

#include "defines.h"

/* ************************************************************************ */
/*                                                                          */
/* ======================== Begin Subroutine ============================== */
/*                                                                          */
/* ************************************************************************ */
void read_atom (void) 
{
	/* ================================================================== */
	/*                                                                    */
	/* Variables, flags, and pointers needed in the subroutine.           */
	/*                                                                    */
	/* ================================================================== */
	int qw=0; //set equal to zero just to prevent a warning.
	char tmp1[5], tmp2[5], tmp3[5],tmp4[5]; 
	int flag1 =0; 
	int flag2 =0; 
	int flag3 =0; 
	int flag4 =0;
	double tmp5;
	char tt[150];
	int rr;
	char *cc;
	FILE *fptr1;
	/* ================================================================== */
	/*                                                                    */
	/* Set the file pointers to the correct input file.                   */
	/*                                                                    */
	/* ================================================================== */

	char fname[100];
	sprintf(fname,"%s",param_file_name);
	if( NULL == (fptr1=fopen(fname,"r")) ) {
		fprintf(stdout,"ERROR: configuration file %s does not exist!!!\n",param_file_name);
		exit(1);
	}

	/* ================================================================== */
	/*                                                                    */
	/* Start reading the file information.                                */
	/*                                                                    */
	/* ================================================================== */
	cc=fgets(tt,150,fptr1);//skip the title line

	/* ----------------------------------------------- */
	/* Read in the atom/site information.              */
	/* ----------------------------------------------- */
	rr=fscanf(fptr1, "%d",   &NUM_ATOM); cc=fgets(tt,150,fptr1);
	atom_type = (struct atomtype*)calloc(NUM_ATOM,sizeof(struct atomtype));
	if(atom_type == NULL){fprintf(stdout, "ERROR: cannot allocate memory for atom_type\n"); exit(11);}

	for (int i=0; i<NUM_ATOM; i++) {
		rr=fscanf(fptr1, "%d",   &atom_type[i].atomid);
		rr=fscanf(fptr1, "%4s",  atom_type[i].name);
		rr=fscanf(fptr1, "%lf\n", &atom_type[i].mass);
	}

	/* ----------------------------------------------- */
	/* Read in the bond parameters.                    */
	/* ----------------------------------------------- */  
	rr=fscanf(fptr1, "%d",   &NUM_BOND); cc=fgets(tt,150,fptr1);
	bond_prop = (struct bondprop*)calloc(NUM_BOND,sizeof(struct bondprop));
	if(bond_prop == NULL){fprintf(stdout, "ERROR: cannot allocate memory for bond_prop\n"); exit(11);}

	/* *********************************************** */
	/* The first loop controls the reading of the      */
	/* bonding atoms and parameters.                   */
	/* *********************************************** */
	for (int i=0; i<NUM_BOND; i++) {
		flag1=flag2=0;

		/* *********************************************** */
		/* Read in the name of the two bonding sites.      */
		/* *********************************************** */
		rr=fscanf(fptr1, "%4s",   tmp1);
		rr=fscanf(fptr1, "%4s",   tmp2);

		/* *********************************************** */
		/* This loop compares the names of the two bonding */
		/* sites with those read in above to assign the    */
		/* correct atomid number.                          */
		/* *********************************************** */
		for (int ii =0; ii<NUM_ATOM; ii++) {
			if (strcmp(tmp1,atom_type[ii].name)==0 ) {
				bond_prop[i].type1=atom_type[ii].atomid;
				flag1 =1;
			}
			if (strcmp(tmp2,atom_type[ii].name)==0 ) {
				bond_prop[i].type2=atom_type[ii].atomid;
				flag2=1;
			}
			if ((flag1*flag2)!=0) break;
		}
		/* *********************************************** */
		/* Read in the harmonic constant and equilibrium   */
		/* bond distance.                                  */
		/* *********************************************** */
		rr=fscanf(fptr1, "%lf",   &bond_prop[i].kbond);
		rr=fscanf(fptr1, "%lf\n", &bond_prop[i].eqbond);
	}

	/* ----------------------------------------------- */
	/* Read in bending(angle) parameters.              */
	/* ----------------------------------------------- */
	rr=fscanf(fptr1, "%d",   &NUM_BEND); cc=fgets(tt,150,fptr1);
	bend_prop = (struct bendprop*)calloc(NUM_BEND,sizeof(struct bendprop));
	if(bend_prop == NULL){fprintf(stdout, "ERROR: cannot allocate memory for bend_prop\n"); exit(11);}  
	/* *********************************************** */
	/* The first loop controls the reading of the      */
	/* bending atoms and parameters.                   */
	/* *********************************************** */
	for (int i=0; i<NUM_BEND; i++) {
		flag1=flag2= flag3=0;

		/* *********************************************** */
		/* Read in the name of the three sites of the      */
		/* bend.                                           */
		/* *********************************************** */
		rr=fscanf(fptr1, "%4s",   tmp1);
		rr=fscanf(fptr1, "%4s",   tmp2);
		rr=fscanf(fptr1, "%4s",   tmp3);

		/* *********************************************** */
		/* This loop compares the names of the three       */
		/* sites involved in the bending with those read   */
		/* in above to assigns the correct atomid number.  */
		/* *********************************************** */
		for (int ii =0; ii<NUM_ATOM; ii++) {
			if (strcmp(tmp1,atom_type[ii].name)==0 ) {
				bend_prop[i].type1=atom_type[ii].atomid;
				flag1 =1;
			}
			if (strcmp(tmp2,atom_type[ii].name)==0 ) {
				bend_prop[i].type2=atom_type[ii].atomid;
				flag2=1;
			}
			if (strcmp(tmp3,atom_type[ii].name)==0 ) {
				bend_prop[i].type3=atom_type[ii].atomid;
				flag3=1;
			}
			if ((flag1*flag2*flag3)!=0) break;
		}
		/* *********************************************** */
		/* Read in the harmonic constant and equilibrium   */
		/* bend angle.                                     */
		/* *********************************************** */
		rr=fscanf(fptr1, "%lf",   &bend_prop[i].kbend);
		rr=fscanf(fptr1, "%lf\n", &bend_prop[i].eqbend);
	}

	/* ----------------------------------------------- */
	/* Read in the dihedral parameters.                */
	/* ----------------------------------------------- */
	rr=fscanf(fptr1, "%d",   &NUM_TORS); cc=fgets(tt,150,fptr1);
	dihedral_prop = (struct dihedralprop*)calloc(NUM_TORS,sizeof(struct dihedralprop));
	if(dihedral_prop == NULL){fprintf(stdout, "ERROR: cannot allocate memory for dihedral_prop\n"); exit(11);}  
	/* *********************************************** */
	/* This first loop controls the reading of the     */
	/* atoms/sites involved in the dihedral angle      */
	/* and the associated parameters.                  */
	/* *********************************************** */
	for (int i=0; i<NUM_TORS; i++) {
		flag1=flag2= flag3=flag4=0;

		/* *********************************************** */
		/* Read in the names of the four sites of the      */
		/* dihedral.                                       */
		/* *********************************************** */
		rr=fscanf(fptr1, "%4s",   tmp1);
		rr=fscanf(fptr1, "%4s",   tmp2);
		rr=fscanf(fptr1, "%4s",   tmp3);
		rr=fscanf(fptr1, "%4s",   tmp4);

		/* *********************************************** */
		/* This loop compares the names of the four        */
		/* sites involved in the dihedral with those read  */
		/* in above to assigns the correct atomid number.  */
		/* *********************************************** */
		for (int ii =0; ii<NUM_ATOM; ii++) {
			if (strcmp(tmp1,atom_type[ii].name)==0 ) {
				dihedral_prop[i].type1=atom_type[ii].atomid;
				flag1 =1;
			}
			if (strcmp(tmp2,atom_type[ii].name)==0 ) {
				dihedral_prop[i].type2=atom_type[ii].atomid;
				flag2=1;
			}
			if (strcmp(tmp3,atom_type[ii].name)==0 ) {
				dihedral_prop[i].type3=atom_type[ii].atomid;
				flag3=1;
			}
			if (strcmp(tmp4,atom_type[ii].name)==0 ) {
				dihedral_prop[i].type4=atom_type[ii].atomid;
				flag4=1;
			}
			if ((flag1*flag2*flag3*flag4)!=0) break;
		}

		/* *********************************************** */
		/* Read in the dihedral force field parameters.    */
		/* *********************************************** */
		rr=fscanf(fptr1, "%lf",   &dihedral_prop[i].kphi);
		rr=fscanf(fptr1, "%lf",   &dihedral_prop[i].nphi);
		rr=fscanf(fptr1, "%lf\n", &dihedral_prop[i].delphi);
	}
	cc=fgets(tt,150,fptr1);

	/* ----------------------------------------------- */
	/* Read in the Lennard Jones parameters.           */
	/* ----------------------------------------------- */    
	/* *********************************************** */
	/* This first loop controls the reading of the     */
	/* atoms/sites.                                    */
	/* *********************************************** */
	for (int i=0; i<NUM_ATOM-1; i++) {

		/* *********************************************** */
		/* Read in the name of the site.                   */
		/* *********************************************** */
		rr=fscanf(fptr1, "%4s",   tmp1);
		/* *********************************************** */
		/* This loop compares the name of the site with    */
		/* the names read in above to assign the proper    */
		/* atomid number.                                  */
		/* *********************************************** */
		for (int ii =0; ii<NUM_ATOM; ii++) {
			if (strcmp(tmp1,atom_type[ii].name)==0 ) {
				qw=atom_type[ii].atomid;
				break;
			}
		}
		/* *********************************************** */
		/* Read in the LJ parameters.                      */
		/* *********************************************** */
		rr=fscanf(fptr1, "%lf",   &epstmp[0][qw]);
		rr=fscanf(fptr1, "%lf\n", &sigtmp[0][qw]);
	}
	/* ----------------------------------------------- */
	/* Read in the improper torsion parameters.        */
	/* ----------------------------------------------- */
	rr=fscanf(fptr1, "%d",   &NUM_IMPR); cc=fgets(tt,150,fptr1);
	improper_prop = (struct improperprop*)calloc(NUM_IMPR,sizeof(struct improperprop));
	if(improper_prop == NULL){fprintf(stdout, "ERROR: cannot allocate memory for improper_prop\n"); exit(11);}  

	/* *********************************************** */
	/* This first loop controls the reading of the     */
	/* atoms/sites involved in the improper angle      */
	/* and the associated parameters.                  */
	/* *********************************************** */
	for (int i=0; i<NUM_IMPR; i++) {
		flag1=flag2= flag3=flag4=0;

		/* *********************************************** */
		/* Read in the names of the four sites of the      */
		/* improper dihedral.                              */
		/* *********************************************** */
		rr=fscanf(fptr1, "%4s",   tmp1);
		rr=fscanf(fptr1, "%4s",   tmp2);
		rr=fscanf(fptr1, "%4s",   tmp3);
		rr=fscanf(fptr1, "%4s",   tmp4);

		/* *********************************************** */
		/* This loop compares the names of the four        */
		/* sites involved in the dihedral with those read  */
		/* in above to assigns the correct atomid number.  */
		/* *********************************************** */
		for (int ii =0; ii<NUM_ATOM; ii++) {
			if (strcmp(tmp1,atom_type[ii].name)==0 ) {
				improper_prop[i].type1=atom_type[ii].atomid;
				flag1 =1;
			}
			if (strcmp(tmp2,atom_type[ii].name)==0 ) {
				improper_prop[i].type2=atom_type[ii].atomid;
				flag2=1;
			}
			if (strcmp(tmp3,atom_type[ii].name)==0 ) {
				improper_prop[i].type3=atom_type[ii].atomid;
				flag3=1;
			}
			if (strcmp(tmp4,atom_type[ii].name)==0 ) {
				improper_prop[i].type4=atom_type[ii].atomid;
				flag4=1;
			}
			if ((flag1*flag2*flag3*flag4)!=0) break;
		}

		/* *********************************************** */
		/* Read in the dihedral force field parameters.    */
		/* *********************************************** */
		rr=fscanf(fptr1, "%lf",   &improper_prop[i].kimpr);
		rr=fscanf(fptr1, "%lf",   &tmp5);
		rr=fscanf(fptr1, "%lf\n", &improper_prop[i].angeq);
	}
	cc=fgets(tt,150,fptr1);
	/* ----------------------------------------------- */
	/* Read in the Lennard Jones special 1-4           */
	/* parameters.                                     */
	/* ----------------------------------------------- */    
	/* *********************************************** */
	/* This first loop controls the reading of the     */
	/* atoms/sites.                                    */
	/* *********************************************** */
	for (int i=0; i<NUM_ATOM-1; i++) {

		/* *********************************************** */
		/* Read in the name of the site.                   */
		/* *********************************************** */
		rr=fscanf(fptr1, "%4s",   tmp1);
		/* *********************************************** */
		/* This loop compares the name of the site with    */
		/* the names read in above to assign the proper    */
		/* atomid number.                                  */
		/* *********************************************** */
		for (int ii =0; ii<NUM_ATOM; ii++) {
			if (strcmp(tmp1,atom_type[ii].name)==0 ) {
				qw=atom_type[ii].atomid;
				break;
			}
		}
		/* *********************************************** */
		/* Read in the LJ parameters.                      */
		/* *********************************************** */
		rr=fscanf(fptr1, "%lf",   &epstmp[1][qw]);
		rr=fscanf(fptr1, "%lf\n", &sigtmp[1][qw]);
	}

	/* ----------------------------------------------- */
	/* Read in the Urey-Bradley 1-3 parameters.        */
	/* ----------------------------------------------- */    

	rr=fscanf(fptr1, "%d",   &NUM_UREY); cc=fgets(tt,150,fptr1);
	urey_prop = (struct ureyprop*)calloc(NUM_UREY,sizeof(struct ureyprop));
	if(urey_prop == NULL){fprintf(stdout, "ERROR: cannot allocate memory for urey_prop\n"); exit(11);}  

	for (int i=0; i<NUM_UREY; i++) {
		flag1=flag2=flag3=0;

		/* *********************************************** */
		/* Read in the name of the three sites of the      */
		/* bend.                                           */
		/* *********************************************** */
		rr=fscanf(fptr1, "%4s",   tmp1);
		rr=fscanf(fptr1, "%4s",   tmp2);
		rr=fscanf(fptr1, "%4s",   tmp3);

		/* *********************************************** */
		/* This loop compares the names of the     two     */
		/* sites involved in the Urey-Bradley 1-3 distance */
		/* with those read in above to assigns the correct */
		/* atomid number.                                  */
		/* *********************************************** */
		for (int ii =0; ii<NUM_ATOM; ii++) {
			if (strcmp(tmp1,atom_type[ii].name)==0 ) {
				urey_prop[i].type1=atom_type[ii].atomid;
				flag1 =1;
			}
			if (strcmp(tmp2,atom_type[ii].name)==0 ) {
				urey_prop[i].type2=atom_type[ii].atomid;
				flag2=1;
			}
			if (strcmp(tmp3,atom_type[ii].name)==0 ) {
				urey_prop[i].type3=atom_type[ii].atomid;
				flag3=1;
			}

			if ((flag1*flag2*flag3)!=0) break;
		}
		/* *********************************************** */
		/* Read in the harmonic constant and equilibrium   */
		/* 1-3 distance.                                   */
		/* *********************************************** */
		rr=fscanf(fptr1, "%lf",   &urey_prop[i].k);
		rr=fscanf(fptr1, "%lf\n", &urey_prop[i].Seq);
	}

	/* ================================================== */
	/* Read in the NBFIX if they are present.             */
	/* ================================================== */
	char key[7] = "NBFIX\n";
	char key2[6] = "NBFIX";
	char test_str[150];
	int ReturnCode;

	ReturnCode = rr=fscanf(fptr1, "%s", test_str);  cc=fgets(tt, 150, fptr1);

	/* This while loops reads the file until the keywork NBFIX is found. */
	while(ReturnCode != EOF){
		if(strcmp(test_str,key)==0 || strcmp(test_str,key2)==0)  break;
		ReturnCode = rr=fscanf(fptr1, "%s", test_str);  cc=fgets(tt, 150, fptr1);;
	}

	/* Now, read the NBFIX                                               */
	if(ReturnCode == EOF)
		NUM_NBFIX = 0;
	else{
		rr=fscanf(fptr1, "%d",   &NUM_NBFIX); cc=fgets(tt,150,fptr1);
		nbfix_prop = (struct nbfixprop*)calloc(NUM_NBFIX,sizeof(struct nbfixprop));
		if(nbfix_prop == NULL){fprintf(stdout, "ERROR: cannot allocate memory for nbfix_prop\n"); exit(11);}
		for (int i=0; i<NUM_NBFIX; i++) {
			flag1=flag2=0;

			/* *********************************************** */
			/* Read in the name of the two types of the        */
			/* nbfix.                                          */
			/* *********************************************** */
			rr=fscanf(fptr1, "%4s",   tmp1);
			rr=fscanf(fptr1, "%4s",   tmp2);

			/* *********************************************** */
			/* This loop compares the names of the two         */
			/* sites involved in the NBFIX                     */
			/* with those read in above to assigns the correct */
			/* atomid number.                                  */
			/* *********************************************** */
			for (int ii =0; ii<NUM_ATOM; ii++){
				if(strcmp(tmp1,atom_type[ii].name)==0 ) {
					nbfix_prop[i].type1=atom_type[ii].atomid;
					flag1 =1;
				}
				if(strcmp(tmp2,atom_type[ii].name)==0 ) {
					nbfix_prop[i].type2=atom_type[ii].atomid;
					flag2=1;
				}
				if ((flag1*flag2*flag3)!=0) break;
			}

			/* *********************************************** */
			/* Read in the eps and sigma.                      */
			/* *********************************************** */
			rr=fscanf(fptr1, "%lf",   &nbfix_prop[i].eps);
			rr=fscanf(fptr1, "%lf\n", &nbfix_prop[i].sig);
		}
	}
	fclose(fptr1);
}

