/* ======================================================================== */
/* read_config.cpp                                                          */
/*                                                              		    */
/*		This subroutine reads in the x,y,z coordinates of the sites from    */
/* the simul#.crd file.  It also does some consistancy checks to make sure  */
/* all the files contain the same number of molecules and sites.  After     */
/* the box information is read in, variables like the box volume, Nose      */
/* Hoover parameters, and SASA parameters are calculated.                   */
/*                                                                          */
/* Note: There is a separate simul#.crd file for each box.                  */
/*                                                                          */
/* ======================================================================== */

#include "defines.h"

/* ================================================================== */
/*                                                                    */
/* Function Prototypes                                                */
/*                                                                    */
/* ================================================================== */
void allocm (void);
void pbc_init (int);
void read_topology(int);

/* ************************************************************************ */
/*                                                                          */
/* ======================== Begin Subroutine ============================== */
/*                                                                          */
/* ************************************************************************ */

void read_config (void) 
{

	sim.NC = 2;
	Nmolall = 2;
	bp[0][0].nbox = 1;
	bp[0][1].nbox = 1;

	mol = (struct mol_info*) calloc(sim.NC,sizeof(struct mol_info));
	if(mol == NULL){ fprintf(stdout, "ERROR: cannot allocate memory for mol\n"); exit(11);}

	mol[0].Nsite = natom_dna.tot/2;
	mol[1].Nsite = natom_dna.tot/2;

	for (int i=0; i<sim.NC; i++) mol[i].Ni = 0;

	for (int k=0; k<sim.NB; k++){
		for (int i=0; i<sim.NC; i++)  mol[i].Ni += bp[k][i].nbox;
	}

	mol[0].Nres = sequence.N;
	mol[1].Nres = sequence.N;

	molec.Nsolute = 2;

	/* ================================================================== */
	/*                                                                    */
	/* Calculate the total number of molecules, in each box by summing up */
	/* the number of each type of molecule in each box.  Also, calculate  */
	/* the total number of sites and residues in each box by multiplying  */
	/* the number of each type of molecule in each box by the number of   */
	/* sites in each molecule and summing them up.                        */
	/*                                                                    */
	/* ================================================================== */
	for (int k=0; k<sim.NB; k++){
		box[k].boxn  = 0;
		box[k].boxns = 0;
		for(int i=0; i<sim.NC; i++) {
			box[k].boxn  += bp[k][i].nbox;// number of molecules in each box
			box[k].boxns += bp[k][i].nbox * mol[i].Nsite;// number of sites in each box
			box[k].boxnres+= bp[k][i].nbox * mol[i].Nres;// number of residues in each box
		}
	}

	/* ----------------------------------------------- */
	/* Check to make sure the total number of          */
	/* molecules is correct. (This check compares      */
	/* numbers from the mol.input file with the        */
	/* number given in simul.input.                    */
	/* ----------------------------------------------- */    
	int qt = 0;
	for (int i=0; i<sim.NC; i++)  
		qt += mol[i].Ni;
	if (qt != Nmolall) {
		fprintf(stdout,"total number of molecules do not match in mol.input!!! (Total number of molecules = %i, sum of individual molecules = %i) \n", Nmolall, qt); exit(1);}

	/* ================================================================== */
	/* Allocate the molec variable and calculate the beginning site of    */
	/* each molecule.                                                     */
	/* ================================================================== */
	molec.fsite = (int*) calloc(box[0].boxn,sizeof(int));
	if(molec.fsite == NULL){ fprintf(stdout, "ERROR: cannot allocate memory for molec.fsite\n"); exit(11);}
	molec.lsite = (int*) calloc(box[0].boxn,sizeof(int));
	if(molec.lsite == NULL){ fprintf(stdout, "ERROR: cannot allocate memory for molec.lsite\n"); exit(11);}
	molec.fres  = (int*) calloc(box[0].boxn,sizeof(int));
	if(molec.fres == NULL){ fprintf(stdout, "ERROR: cannot allocate memory for molec.fres\n"); exit(11);}
	molec.lres  = (int*) calloc(box[0].boxn,sizeof(int));
	if(molec.lres == NULL){ fprintf(stdout, "ERROR: cannot allocate memory for molec.lres\n"); exit(11);}

	int nmolec = 0;
	for(int i=0; i<sim.NC; i++){
		for(int j=0; j<bp[0][i].nbox; j++){
			if(i==0 && j==0){
				molec.fsite[0] = 0;
				molec.lsite[0] = mol[0].Nsite;
				molec.fres[0]  = 0;
				molec.lres[0]  = mol[0].Nres;
			}
			else{
				molec.fsite[nmolec] = molec.lsite[nmolec-1];
				molec.lsite[nmolec] = molec.fsite[nmolec] +mol[i].Nsite;
				molec.fres[nmolec]  = molec.lres[nmolec-1];
				molec.lres[nmolec]  = molec.lres[nmolec-1]+mol[i].Nres;
			}
			nmolec++;
		}
	}

	/* ================================================================== */
	/* Now that the box numbers are known, allocate the needed memory.    */
	/* ================================================================== */
	allocm();

	/* ================================================================== */
	/* Read in the connectivity.                                          */
	/* ================================================================== */
	for(int k=0; k<sim.NB; k++) read_topology(k);

	/* ================================================================== */
	/* Now, read in the x,y,z coordinates from the simul#.crd file.       */
	/* ================================================================== */
	for(int k=0; k<sim.NB; k++) {

		/* ----------------------------------------------- */
		/* Variables needed to read the files.             */
		/* ----------------------------------------------- */    

		/* ----------------------------------------------- */
		/* Read in the site name and residue number of     */
		/* each site.  Compare them to those read in       */
		/* read_topology().  If they are different, exit.  */
		/* ----------------------------------------------- */    
		for (int i=0; i<box[k].boxns; i++){
			/* ----------------------------------------------- */
			/* Read in the site x,y,z coordinates.             */
			/* ----------------------------------------------- */    

			atom[k][i].x = atom_dna[i].x;
			atom[k][i].y = atom_dna[i].y;
			atom[k][i].z = atom_dna[i].z;

			// NOTE: Rounding coordinates to match the PDB file format
			atom[k][i].x = ROUND(atom[k][i].x,3);
			atom[k][i].y = ROUND(atom[k][i].y,3);
			atom[k][i].z = ROUND(atom[k][i].z,3);

		}// end i
	}// end k

	for(int k=0; k<sim.NB; k++){
		box[k].rc2 = sim.rc * sim.rc;
		box[k].rc4 = box[k].rc2 * box[k].rc2;	
	}

}
