/* ======================================================================== */
/* setlj.cpp                                                                */
/*                                                                          */
/*		This subroutine sets the Lennard-Jones parameters for each          */
/* interaction. It applies the Lorentz-Bertheloth mixing rules and stores   */
/* the parameters for the interaction of particles types a and b in         */
/* ljset[a][b].  This the the structure used in cnonbond and                */
/* cononbond_nblist to calculate the LJ interactions.  The hash table,      */
/* used for the excluded interactions, is also set up here.                 */
/*                                                                          */
/* Passed Parameters:                                                       */
/*                                                                          */
/* ======================================================================== */
#include "defines.h"

/* ************************************************************************ */
/*                                                                          */
/* ======================== Begin Subroutine ============================== */
/*                                                                          */
/* ************************************************************************ */
void setlj (void)
{
	/* ================================================================== */
	/* Loop around number of site types interactions and apply the mixing */
	/* rules.                                                             */
	/* ================================================================== */
	for(int a=0; a<NUM_ATOM; a++) {
		for(int b=0; b<NUM_ATOM; b++) {
			ljset[atom_type[a].atomid][atom_type[b].atomid].eps = 4.184 * sqrt(epstmp[0][atom_type[a].atomid] * epstmp[0][atom_type[b].atomid]);
			ljset[atom_type[a].atomid][atom_type[b].atomid].sig = 0.5 * (2.0 * sigtmp[0][atom_type[a].atomid] + 2.0 * sigtmp[0][atom_type[b].atomid]);//the 2.0 comes from the fact that CHARMM lists sig/2 
			ljset[atom_type[a].atomid][atom_type[b].atomid].id = 0; //0 means there are no special nbfix, native contacts,      etc.

		}
	}

	/* ================================================================== */
	/* If a and b are part of a NBFIX, use the specified interaction      */
	/* parameters.                                                        */
	/* ================================================================== */
	for (int n=0; n<NUM_NBFIX; n++){
		ljset[nbfix_prop[n].type1][nbfix_prop[n].type2].eps = nbfix_prop[n].eps*-4.184;
		ljset[nbfix_prop[n].type2][nbfix_prop[n].type1].eps = nbfix_prop[n].eps*-4.184;    
		ljset[nbfix_prop[n].type1][nbfix_prop[n].type2].sig = nbfix_prop[n].sig; 
		ljset[nbfix_prop[n].type2][nbfix_prop[n].type1].sig = nbfix_prop[n].sig; 
	}

	/* ================================================================== */
	/*                                                                    */
	/* Set up hash table of excluded interactions.                        */
	/*                                                                    */
	/* ================================================================== */
	int maxn=0;
	hashsize=1000;
	for(int k=0; k<sim.NB; k++){
		table[k] = (struct hashlist*) calloc (NPRIME+1, sizeof(struct hashlist));
		if(table[k] == NULL){ fprintf(stdout, "ERROR: cannot allocate memory for table\n"); exit(1);}

		for(int i=0; i<(NPRIME+1); i++){
			table[k][i].a = (int*) calloc(hashsize, sizeof(int));
			table[k][i].b = (int*) calloc(hashsize, sizeof(int));
			table[k][i].p = (int*) calloc(hashsize, sizeof(int));
			table[k][i].n = 0;
		}//for i
		for(int i=0; i<xintN[k]; i++){
			int a = xint[k][i].a;
			int b = xint[k][i].b;
			unsigned long int index = (a*b) % NPRIME;
			if(table[k][index].n == 0){
				table[k][index].a[0] = a;
				table[k][index].b[0] = b;
				table[k][index].p[0] = i;
			}
			else{
				int qa = table[k][index].n;
				table[k][index].a[qa] = a;
				table[k][index].b[qa] = b;
				table[k][index].p[qa]  = i;
			}
			table[k][index].n++;
			if( table[k][index].n>maxn)
			{
				maxn = table[k][index].n;
				if(maxn > hashsize){
					fprintf(stdout,"The hash table isn't big enough. Increase 'hashsize' in setlj.cpp\n");
				}
			}
		}//for i
	}//for k

}
